/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.admin;

import java.net.*;
import java.io.*;
import java.text.*;
import java.util.*;

import org.w3c.dom.*;

import org.log4j.Category;

import de.hendriklipka.databus.*;

import de.hendriklipka.util.*;

import de.hendriklipka.gcp.exception.*;
import de.hendriklipka.gcp.module.admin.handler.*;
import de.hendriklipka.gcp.event.*;

/**
  NetworkControl is a component, which allows the remote control of the running
  program a'la QControl.
  Somebody can connect via telnet and can send commands to the handler (these
  commands start with '/') and to the running programm.
  For every connection a single thread is started.
  execution of commands is done via command handler modules.
*/
public class NetworkControl implements DataBusComponent, DataBusConsumer, DataBusProducer, Runnable
{
    static Category CAT = Category.getInstance(NetworkControl.class.getName());

    protected DataBus _bus=null;
    
    protected boolean _startShutdown=false;
    protected Thread _myThread=null;
    
    protected Hashtable _users=new Hashtable();
    protected Hashtable _rolesToHandlers=new Hashtable();
    protected Hashtable _rolesForOutput=new Hashtable();
    protected Hashtable _rolesForCommands=new Hashtable();
    
    protected ServerSocket _serverSocket=null;
    
    protected final int LOG_ERROR=1;
    protected final int LOG_LOGIN=2;
    protected final int LOG_COMMAND=3;
    protected final int LOG_RESULT=4;
    protected final int LOG_DEBUG=5;
    
    protected Options _options=null;
    
    protected PrintWriter _logWriter=null;
    protected DateFormat _formatter=DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT);
    
    protected class UserDef
    {
        public String _login=null;
        public String _password=null;
        public String _role=null;
    }
    
    protected class Options
    {
        public int _port=26000;
        public int _passwordTries=1;
        public String _logfileName="Control.log";
        public String _logLevelStr=null;
        public int _logLevel=0;
        public int _timeout=1;
        public String _banner="Logalyzer (C)1998-2000 by Hendrik Lipka";
        public String _linefeed="\r\n";
        public String _prefix="/";
        public String _commandRoles=null;
        public String _outputRoles=null;
        
        public void handleUsers(Node node, Properties data)
        {
            _users.clear();
            
            CAT.info("<users>");
            NodeList nl2=node.getChildNodes();
            for (int j=0;j<nl2.getLength();j++)
            {
                Node node2=nl2.item(j);
                if (node2.getNodeType()==node.ELEMENT_NODE)
                {
                    CAT.info("<user>");
                    if (node2.getNodeName().equals("user"))
                        createUser(node2);
                }
            }
        }
        
        public void handleHandlers(Node node, Properties data)
        {
            _rolesToHandlers.clear();

            NodeList nl2=node.getChildNodes();
            for (int j=0;j<nl2.getLength();j++)
            {
                Node node2=nl2.item(j);
                if (node2.getNodeType()==node.ELEMENT_NODE)
                {
                    if (node2.getNodeName().equals("handler"))
                        createHandler(node2);
                }
            }
        }
        
        public void postProcess()
        {
            if (_linefeed.equals("CR"))
            {
                _linefeed="\r";
            }
            else if (_linefeed.equals("CRLF"))
            {
                _linefeed="\r\n";
            }
            else if (_linefeed.equals("LF"))
            {
                _linefeed="\n";
            }
            else if (_linefeed.equals("LFCR"))
            {
                _linefeed="\n\r";
            }
            else
            {
                _linefeed="";
            }
            
            CAT.debug("logLevel="+_logLevelStr);
            if (_logLevelStr.equals("ERROR"))
            {
                _logLevel=1;
            }
            else if (_logLevelStr.equals("LOGIN"))
            {
                _logLevel=2;
            }
            else if (_logLevelStr.equals("COMMAND"))
            {
                _logLevel=3;
            }
            else if (_logLevelStr.equals("RESULT"))
            {
                _logLevel=4;
            }
            else if (_logLevelStr.equals("DEBUG"))
            {
                _logLevel=5;
            }
            else
            {
                _logLevel=0;
            }
            CAT.debug("logLevel="+_logLevel);
            
            _rolesForCommands.clear();
            String[] sa=Util.stringToArray(_commandRoles);
            for (int j=0;j<sa.length;j++)
            {
                _rolesForCommands.put(sa[j],sa[j]);
            }
            
            _rolesForOutput.clear();
            sa=Util.stringToArray(_outputRoles);
            for (int j=0;j<sa.length;j++)
            {
                _rolesForOutput.put(sa[j],sa[j]);
            }
        }
    }

    protected String[][] _rules={
                        {"_port","port"},
                        {"_passwordTries","tries"},
                        {"_logfileName","logfile"},
                        {"_logLevelStr","loglevel"},
                        {"_timeout","timeout"},
                        {"_banner","banner"},
                        {"_linefeed","linefeed"},
                        {"handleUsers()","users"},
                        {"handleHandlers()","handlers"},
                        {"_prefix","commandprefix"},
                        {"_outputRoles","roles-for-output"},
                        {"_commandRoles","roles-for-direct-commands"}};
    protected String[] _checkRules={
                        "_port",
                        "_passwordTries",
                        "_logfileName",
                        "_logLevelStr",
                        "handleUsers()",
                        "handleHandlers()",
                        "_prefix"};
                        
    public void init(DataBus bus, Node configNode) throws Exception
    {
        _bus=bus;
        _options= new Options();

        try
        {
            XMLUtil.readXMLConfigNode(configNode,_rules,_checkRules,null,_options);
            _options.postProcess();
            _logWriter=new PrintWriter(new FileWriter(_options._logfileName));
                _serverSocket=new ServerSocket(_options._port,1024);
                _serverSocket.setSoTimeout(1000*_options._timeout);
        }
        catch (Exception ex)
        {
            CAT.error("while opening log file: ",ex);
            throw ex;
        }
    }
      
    public void createUser(Node node)
    {
        try
        {
            UserDef ud=new UserDef();
            NamedNodeMap nnm=node.getAttributes();
            Node attr=nnm.getNamedItem("login");
            ud._login=attr.getNodeValue();
            attr=nnm.getNamedItem("role");
            ud._role=attr.getNodeValue();
            attr=nnm.getNamedItem("password");
            ud._password=attr.getNodeValue();
            _users.put(ud._login,ud);
        }
        catch (Exception ex)
        {
            CAT.error("while reading user data: ",ex);
        }
    }
    
    public void createHandler(Node node)
    {
        NamedNodeMap nnl=node.getAttributes();
        Node attr=nnl.getNamedItem("class");
        if (null==attr)
        {
            CAT.error("handler w/o class");
            return;
        }
        String cname=attr.getNodeValue();

        attr=nnl.getNamedItem("role");
        if (null==attr)
        {
            CAT.error("handler w/o role");
            return;
        }
        String roles=attr.getNodeValue();
        String[] sa=Util.stringToArray(roles);        

        try
        {
            Class c=Class.forName(cname);
            Object o=c.newInstance();
            CommandHandler ch=(CommandHandler)o;
            ch.init(_bus,node);
            for (int i=0;i<sa.length;i++)
            {
                addHandlerToRole(sa[i],ch);
            }
        }
        catch (Exception ex)
        {
            CAT.error("while creating command handler: ",ex);
        }
    }
        
    public void addHandlerToRole(String role, CommandHandler ch)
    {
        Object o=_rolesToHandlers.get(role);
        if (null==o)
        {
            Vector v=new Vector();
            v.addElement(ch);
            _rolesToHandlers.put(role,v);
        }
        else
        {
            Vector v=(Vector)o;
            v.addElement(ch);
        }
    }
    
    public void run() 
    {
        _myThread=Thread.currentThread();
        int conCount=1;
        String conPrefix="con";
        try
        {
            while(!_startShutdown)
            {
                try
                {
                    Socket mySocket=_serverSocket.accept();
                    if (_startShutdown)
                    {
                        break;
                    }
                    if (null==mySocket)
                        continue;
                    ConnectHandler ch=null;
                    try
                    {
                        CAT.debug("creating connect handler");
                        String name=conPrefix+Integer.toString(conCount);
                        conCount++;
                        ch=new ConnectHandler(mySocket,name);
                        ch.init(_bus);
                        ch.start();
                    }
                    catch (Throwable ex)
                    {
                        CAT.error("while creating connect handler: ",ex);
                        mySocket.close();
                        continue;
                    }
                }
                catch (IOException ex)
                {
                    synchronized (this)
                    {
                        try
                        {
                            wait(50);
                        }
                        catch (InterruptedException ex2) {}
                    }
                }
            }
        }
        finally
        {
            try
            {
                _serverSocket.close();
                _serverSocket=null;
            }
            catch (IOException ex) {}
        }
    }
      
    protected void logMessage(int level, String handlerName, String message)
    {
        logMessage(level,handlerName,message,null);
    }
    
    protected void logMessage(int level, String handlerName, String message, Throwable ex)
    {
        if (null!=_logWriter)
        {
            if (level<=_options._logLevel)
            {
                String time=_formatter.format(new Date(System.currentTimeMillis()));
                _logWriter.println(time+" "+handlerName+" : "+message);
                if (null!=ex)
                {
                    _logWriter.print(time+" "+handlerName+" : exception : ");
                    ex.printStackTrace(_logWriter);
                }
                _logWriter.flush();
            }
        }
    }

    /**
        direct commands (commands which do not start with the command prefix) are
        handled by this method (and not from a command handler)
    */    
    protected String handleDirectCommand(String cmd)
    {
        GameControlEvent gce=new GameControlEvent();
        gce._type=GameControlEvent.USER_DEFINED;
        gce._param1=cmd;
        Object o=_bus.sendResponseData(gce,GameControlEvent.NAME,this);
        String answer="";
        try
        {
            answer=((GameControlResponse)o)._response;
        }
        catch (Exception ex) 
        {
            CAT.error(ex);
        }
        if (null==answer) 
            answer="";
        return answer;
    }
        
    public synchronized void dataAvailable(de.hendriklipka.databus.DataBusEvent o)
    {
    }
      
    public void start()
    {
        new Thread(this).start();
    }
            
    public void shutdown()
    {
        _startShutdown=true;
    }
            
    public void deInitialize()
    {
    }
            
    public void reconfig(Node configNode) throws Exception
    {
        Options options= new Options();

        try
        {
            XMLUtil.readXMLConfigNode(configNode,_rules,_checkRules,null,options);
            options.postProcess();
            if (!options._logfileName.equals(_options._logfileName))
            {
                if (null!=_logWriter)
                    _logWriter.close();
                _logWriter=new PrintWriter(new FileWriter(options._logfileName));
            }
            if (options._port!=_options._port)
            {
                _serverSocket=new ServerSocket(_options._port,1024);
                _serverSocket.setSoTimeout(1000*_options._timeout);
            }
            _options=options;
        }
        catch (Exception ex)
        {
            CAT.error("while opening log file: ",ex);
            throw ex;
        }
    }
            
    public void preStartup() throws Exception
    {
    }
    
    public void postStartup() throws Exception
    {
    }
    
    /**
        handles a connection to a client - one thread per connection
    */
    class ConnectHandler extends Thread implements DataBusConsumer
    {
        protected String _name=null;
        protected Socket _mySocket=null;
        protected boolean _isShutdown=false;
        protected PrintWriter _bw=null;
        
        protected UserDef _currentUser=null;
        
        protected CommandHandler _defaultHandler=null;
        protected CommandHandler[] _handlers=new CommandHandler[1];
        
        public ConnectHandler(Socket s, String name)
        {
            _mySocket=s;
            _name=name;
            CAT.debug(s.getInetAddress().toString());
        }

        public void init(DataBus bus)
        {
            _bus=bus;
            _bus.addConsumer(this,InputEvent.NAME);
            _handlers[0]=new AuthCommandHandler(this);
            _handlers[0].init(bus,null);
            _defaultHandler=_handlers[0];
        }
            
        protected CommandHandler[] getHandlersForRole(String role)
        {
            Vector v=(Vector)_rolesToHandlers.get(role);
            if (null==v)
            {
                CAT.error("no handlers found for role "+role);
                logMessage(LOG_ERROR,_name,"no handlers found for role "+role);
                return null;
            }
            CommandHandler[] cha=new CommandHandler[v.size()+1];
            for (int i=0;i<cha.length-1;i++)
            {
                cha[i]=(CommandHandler)v.elementAt(i);
            }
            cha[cha.length-1]=_defaultHandler;
            return cha;
        }

        /**
            handles login
        */        
        protected UserDef loginUser(String user, String password)
        {
            logMessage(LOG_LOGIN,_name,"trying to login user "+user+"/"+password);
            UserDef ud=(UserDef)_users.get(user);
            if (null==ud)
            {
                CAT.error("user "+user+" not found");
                logMessage(LOG_LOGIN,_name,"user "+user+" not found");
                return null;
            }
            if (!password.equals(ud._password))
            {
                CAT.error("password for user "+user+" does not match");
                logMessage(LOG_LOGIN,_name,"password for user "+user+" does not match");
                return null;
            }
            return ud;
        }

        /**
            handles commands
            tries to find all command handlers for the current user
            then they get tha command for handling
            after the first handler which returns isHandled()==true  the excution is finished
        */    
        protected String handleCommand(String cmd)
        {
            logMessage(LOG_COMMAND,_name,"command: "+cmd);
            if (cmd.startsWith(_options._prefix))
            {
                cmd=cmd.substring(_options._prefix.length());
                for (int i=0;i<_handlers.length;i++)
                {
                    CommandResponse res=_handlers[i].handleCommand(cmd);
                    if ((null!=res)&&(res.isHandled()))
                    {
                        logMessage(LOG_RESULT,_name,"result: "+res.getResponse());
                        return res.getResponse();
                    }
                }
                logMessage(LOG_COMMAND,_name,"unknown command");
                return "unknown command";
            }
            else
            {
                // check if the current user is allowd to send direct commands
                if (null!=_currentUser)
                {
                    Object o=_rolesForCommands.get(_currentUser._role);
                    if (null!=o)
                    {
                        String s=handleDirectCommand(cmd);
                        logMessage(LOG_RESULT,_name,"result: "+s);
                        return s;
                    }
                }
                return "you are not allowed to use direct commands";
            }
        }
        
        public void run()
        {
            try
            {
                String ls="\n\r";
                CAT.info("starting connect handler");
                try
                {
                    InputStream is=_mySocket.getInputStream();
                    OutputStream os=_mySocket.getOutputStream();
                    _bw=new PrintWriter(new OutputStreamWriter(os),true);
                    _bw.print(_options._banner+_options._linefeed);     
                    _bw.flush();
                    boolean fini=false;
                    while (!fini)
                    {
                        StringBuffer sb=new StringBuffer();
                        while (true)
                        {
                            int i=is.read();
                            if (_isShutdown)
                            {
                                return;
                            }
                            if (-1==i)
                            {
                                fini=true;
                                break;
                            }
                            // handle end of line
                            if (-1!=ls.indexOf((char)i))
                            {
                                if (0==sb.length())
                                    continue;
                                break;
                            }
                            _bw.write(i);
                            _bw.flush();
                            sb.append((char)i);
                        }
                        if (0==sb.length())
                            continue;
                        // handle the command
                        _bw.print(_options._linefeed);
                        _bw.flush();
                        String command=sb.toString();
                        CAT.debug("command: <"+command+">");
                        String r=handleCommand(command);
                        if (null==r)
                        {
                            fini=true;
                            _bw.print("connection terminated by server"+_options._linefeed);
                            _bw.flush();
                            break;
                        }
                        if (0!=r.length())
                        {
                            _bw.print(r+_options._linefeed);
                            _bw.flush();
                        }
                        if (0==_options._passwordTries)
                        {
                            _bw.print("connection terminated by server"+_options._linefeed);
                            fini=true;
                        }
                    }
                }
                catch (Exception ex)
                {
                    CAT.error(ex);
                }
                finally
                {
                    try
                    {
                        CAT.debug("closing the socket: "+_mySocket.toString());
                        _mySocket.close();
                    }
                    catch (IOException ex) {}
                }
            }
            finally
            {
                _bus.removeConsumer(this);
            }
        }

        public synchronized void dataAvailable(de.hendriklipka.databus.DataBusEvent data) 
        {
            if (null!=_currentUser)
            {
                Object o=_rolesForOutput.get(_currentUser._role);
                if (null!=o)
                {
                    InputEvent ie=(InputEvent)data._data;
                    _bw.print(ie._line);
                    _bw.print(_options._linefeed);
                    _bw.flush();
                }
            }
        }
        
        public String[] listenTo()
        {
            String[] lt={InputEvent.NAME};
            return lt;
        }
    } // class ConnectHandler

    /**
        internal command handler - handles authentication and exit
    */      
    protected class AuthCommandHandler extends CommandHandler
    {
        protected ConnectHandler _handler=null;
        
        public AuthCommandHandler(ConnectHandler handler)
        {
            _handler=handler;
        }
        
        public String[] getHandlerCommands()
        {
            String[] sa= {"login","quit"};
            return sa;
        }
        
        public CommandResponse handleCommand(String cmd)
        { 
            if (cmd.startsWith("login"))
            {
                try
                {
                    StringTokenizer st=new StringTokenizer(cmd," \t");
                    String s1=(String)st.nextElement();
                    String name=(String)st.nextElement();
                    String pwd=(String)st.nextElement();
                    UserDef ud=_handler.loginUser(name, pwd);
                    if (null==ud)
                    {
                        return new CommandResponse("can't login this user");
                    }
                    CommandHandler[] cha=_handler.getHandlersForRole(ud._role);
                    if (null==cha)
                    {
                        return new CommandResponse("no known commands for this user");
                    }
                    _handler._handlers=cha;
                    _handler._currentUser=ud;
                    return new CommandResponse("ok");
                }
                catch (Exception ex)
                {
                    NetworkControl.CAT.debug("login: <"+cmd+">",ex);
                    return new CommandResponse("usage: login name password");
                }
            }
            else if (cmd.equals("quit"))
            {
                _handler._currentUser=null;
                return new CommandResponse(null);
            }
            return new CommandResponse();
        }
    } // class AuthCommandHandler
} 
