/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.control;

import java.util.*;

import org.w3c.dom.*;

import org.log4j.*;

import de.hendriklipka.databus.*;
import de.hendriklipka.util.*;

import de.hendriklipka.gcp.exception.*;
import de.hendriklipka.gcp.event.*;

public class CommandScheduler implements DataBusComponent, DataBusProducer
{
    static Category CAT = Category.getInstance(CommandScheduler.class.getName());

    protected DataBus _bus=null;

    protected SortedSet _queue=new TreeSet();
    protected boolean _shutDown=false;
    protected Object _waitObj=new Object();

    public void init(DataBus bus, Node configNode) throws java.lang.Exception
    {
        _bus=bus;
        reconfig(configNode);
    }

    public void preStartup() throws java.lang.Exception
    {
    }

    public void start() throws java.lang.Exception
    {
    }

    // take all events, schedule them and put them into the queue
    // we use a copy of the queue so we don't mix scheduled and unscheduled events
    public void postStartup() throws java.lang.Exception
    {
        List events=new ArrayList(_queue);
        _queue.clear();
        for (int i=0;i<events.size();i++)
        {
            ActionEvent event=(ActionEvent)events.get(i);
            calculateNextRun(event);
            if (-1!=event._nextRun)
            {
                _queue.add(event);
            }
        }

        Thread t=new SchedulerThread();
        t.setPriority(t.getPriority()+1);
        t.start();
    }

    // calculate when to run the event the next time
    protected void calculateNextRun(ActionEvent event)
    {
        try
        {
            CAT.debug("calculateNextRun("+event._id+")");
            // if we have an event with interval, calculation is simple...
            if (-1!=event._interval)
            {
//                CAT.debug("adding event for interval "+event._interval);
                event._nextRun=System.currentTimeMillis()+event._interval*1000;
                return;
            }
            // Logic is as follows:
            // calculate the current time and get all fields
            // - starting from year upwards, search through all field arrays for the first value
            //  greater than or equal the current value
            // - if only a value greater than the current one is found, set all other fields to their
            //  lowest possible value
            // - otherwise, continue with the next lower key
            // - if for a key no value greater than or equal to the current value can be found,
            // - use the lowest possible value and add 1 to the next higher field (and upwards, if needed)
            // there is a special handling for specified day-of-week:
            // - when searching for a possible day-of-month, we need to look also at the possible
            //  days-of-week
            // - NOTE: this can lead to wrapping the month field!
            // as the arrays containing the possible values are sorted, for searching we can use
            // binary-search (Arrays.binarySearch(array,key))
            
            // get current time + 0.5 seconds
            Calendar cal=Calendar.getInstance();
            cal.setTime(new Date(System.currentTimeMillis()+1000));
            int year=cal.get(Calendar.YEAR);
            int month=cal.get(Calendar.MONTH);
            int day=cal.get(Calendar.DAY_OF_MONTH);
            int dayOfWeek=cal.get(Calendar.DAY_OF_WEEK);
            int hour=cal.get(Calendar.HOUR_OF_DAY);
            int minute=cal.get(Calendar.MINUTE);
            int second=cal.get(Calendar.SECOND);
            
            int pos=Arrays.binarySearch(event._years,year);
            CAT.debug("year - pos="+pos);
            if (0>pos) // year not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._years.length) // no higher value for year
                {
                    event._nextRun=-1;
                    CAT.debug("no scheduling possible (year)");
                    return;
                }
                else
                {
                    event._currYearIdx=pos;
                    event._currMonthIdx=0;
                    event._currDayIdx=0;
                    event._currHourIdx=0;
                    event._currMinIdx=0;
                    event._currSecIdx=0;
                    if (!checkDayOfWeek(event))
                    {
                        event._nextRun=-1;
                        CAT.debug("no scheduling possible (day@month)");
                        return;
                    }
                    event._nextRun=calculateTime(event);
                    CAT.debug("finished@year");
                    return;
                }
            }
            
            // year was found, so we need to check for the month
            event._currYearIdx=pos;
            pos=Arrays.binarySearch(event._months,month);
            CAT.debug("month - pos="+pos);
            if (0>pos) // month not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._months.length) // no higher value for month
                {
                    CAT.debug("next year@month");
                    event._currMonthIdx=0;
                    event._currYearIdx++;
                    if (event._currYearIdx>=event._years.length) // no higher value for year
                    {
                        event._nextRun=-1;
                        CAT.debug("no scheduling possible (month)");
                        return;
                    }
                }
                else
                {
                    event._currMonthIdx=pos;
                }
                event._currDayIdx=0;
                event._currHourIdx=0;
                event._currMinIdx=0;
                event._currSecIdx=0;
                if (!checkDayOfWeek(event))
                {
                    event._nextRun=-1;
                    CAT.debug("no scheduling possible (day@month)");
                    return;
                }
                event._nextRun=calculateTime(event);
                CAT.debug("finished@month");
                return;
            }
            
            // month was found, so we need to check for the day
            event._currMonthIdx=pos;
            pos=Arrays.binarySearch(event._days,day);
            CAT.debug("day - pos="+pos);
            if (0>pos) // day not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._days.length) // no higher value for day
                {
                    event._currDayIdx=0;
                    event._currMonthIdx++;
                    CAT.debug("next month@day");
                    if (event._currMonthIdx>=event._months.length) // no higher value for month
                    {
                        event._currMonthIdx=0;
                        event._currYearIdx++;
                        CAT.debug("next year@day");
                        if (event._currYearIdx>=event._years.length) // no higher value for year
                        {
                            event._nextRun=-1;
                            CAT.debug("no scheduling possible (year@day)");
                            return;
                        }
                    }
                }
                else
                {
                    event._currDayIdx=pos;
                }
                event._currHourIdx=0;
                event._currMinIdx=0;
                event._currSecIdx=0;
                if (!checkDayOfWeek(event))
                {
                    event._nextRun=-1;
                    CAT.debug("no scheduling possible (day@day)");
                    return;
                }
                event._nextRun=calculateTime(event);
                CAT.debug("finished@day");
                return;
            }
            
            // day was found, so we need to check for the hour
            event._currDayIdx=pos;
            pos=Arrays.binarySearch(event._hours,hour);
            CAT.debug("hour - pos="+pos);
            if (0>pos) // hour not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._hours.length) // no higher value for hour
                {
                    event._currHourIdx=0;
                    event._currDayIdx++;
                    CAT.debug("next day@hour");
                    if (event._currDayIdx>=event._days.length) // no higher value for day
                    {
                        event._currDayIdx=0;
                        event._currMonthIdx++;
                        CAT.debug("next month@hour");
                        if (event._currMonthIdx>=event._months.length) // no higher value for month
                        {
                            event._currMonthIdx=0;
                            event._currYearIdx++;
                            CAT.debug("next year@hour");
                            if (event._currYearIdx>=event._years.length) // no higher value for year
                            {
                                event._nextRun=-1;
                                CAT.debug("no scheduling possible (year@hour)");
                                return;
                            }
                        }
                    }
                    // we have changed the day, so we need to check it again
                    if (!checkDayOfWeek(event))
                    {
                        event._nextRun=-1;
                        CAT.debug("no scheduling possible (day@hour)");
                        return;
                    }
                }
                else
                {
                    event._currHourIdx=pos;
                }
                event._currMinIdx=0;
                event._currSecIdx=0;
                event._nextRun=calculateTime(event);
                CAT.debug("finished@hour");
                return;
            }
            
            // hour was found, so we need to check for the minute
            event._currHourIdx=pos;
            pos=Arrays.binarySearch(event._minutes,minute);
            CAT.debug("minute - pos="+pos);
            if (0>pos) // minute not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._minutes.length) // no higher value for minute
                {
                    event._currMinIdx=0;
                    event._currHourIdx++;
                    CAT.debug("next hour@min");
                    if (event._currHourIdx>=event._hours.length) // no higher value for hour
                    {
                        event._currHourIdx=0;
                        event._currDayIdx++;
                        CAT.debug("next day@min");
                        if (event._currDayIdx>=event._days.length) // no higher value for day
                        {
                            event._currDayIdx=0;
                            event._currMonthIdx++;
                            CAT.debug("next month@min");
                            if (event._currMonthIdx>=event._months.length) // no higher value for month
                            {
                                event._currMonthIdx=0;
                                event._currYearIdx++;
                                CAT.debug("next year@min");
                                if (event._currYearIdx>=event._years.length) // no higher value for year
                                {
                                    event._nextRun=-1;
                                    CAT.debug("no scheduling possible (year@min)");
                                    return;
                                }
                            }
                        }
                        // we have changed the day, so we need to check it again
                        if (!checkDayOfWeek(event))
                        {
                            event._nextRun=-1;
                            CAT.debug("no scheduling possible (day@min)");
                            return;
                        }
                    }
                }
                else
                {
                    event._currMinIdx=pos;
                }
                event._currSecIdx=0;
                event._nextRun=calculateTime(event);
                CAT.debug("finished@min");
                return;
            }
            
            // minute was found, so we need to check for the second
            event._currMinIdx=pos;
            pos=Arrays.binarySearch(event._seconds,second);
            CAT.debug("second - pos="+pos);
            if (0>pos) // second not found
            {
                pos=(-pos)-1; // calculate next higher position
                if (pos>=event._seconds.length) // no higher value for second
                {
                    event._currSecIdx=0;
                    event._currMinIdx++;
                    CAT.debug("next min@sec");
                    if (event._currMinIdx>=event._minutes.length) // no higher value for minute
                    {
                        event._currMinIdx=0;
                        event._currHourIdx++;
                        CAT.debug("next hour@sec");
                        if (event._currHourIdx>=event._hours.length) // no higher value for hour
                        {
                            event._currHourIdx=0;
                            event._currDayIdx++;
                            CAT.debug("next day@sec");
                            if (event._currDayIdx>=event._days.length) // no higher value for day
                            {
                                event._currDayIdx=0;
                                event._currMonthIdx++;
                                CAT.debug("next month@sec");
                                if (event._currMonthIdx>=event._months.length) // no higher value for month
                                {
                                    event._currMonthIdx=0;
                                    event._currYearIdx++;
                                    CAT.debug("next year@sec");
                                    if (event._currYearIdx>=event._years.length) // no higher value for year
                                    {
                                        event._nextRun=-1;
                                        CAT.debug("no scheduling possible (year@sec)");
                                        return;
                                    }
                                }
                            }
                            // we have changed the day, so we need to check it again
                            if (!checkDayOfWeek(event))
                            {
                                event._nextRun=-1;
                                CAT.debug("no scheduling possible (day@sec)");
                                return;
                            }
                        }
                    }
                }
                else
                {
                    event._currSecIdx=pos;
                }
                event._nextRun=calculateTime(event);
                CAT.debug("finished@sec");
                return;
            }
            event._currSecIdx=pos;
            event._nextRun=calculateTime(event);
            CAT.debug("event scheduled: "+event);
        }
        catch (Exception ex)
        {
            CAT.error("error during calculation of next run",ex);
        }
    }

    protected boolean checkDayOfWeek(ActionEvent event)
    {
        CAT.debug("checkDayOfWeek - "+event);
        Calendar cal=Calendar.getInstance();
        cal.setLenient(true);
        cal.set(event._years[event._currYearIdx],
                    event._months[event._currMonthIdx],
                    event._days[event._currDayIdx],
                    event._hours[event._currHourIdx],
                    event._minutes[event._currMinIdx],
                    event._seconds[event._currSecIdx]
                );
        int dow=cal.get(Calendar.DAY_OF_WEEK);
        int pos=Arrays.binarySearch(event._daysOfWeek,dow);
        if ((0<=pos)&&(event._days[event._currDayIdx]<=cal.getActualMaximum(Calendar.DAY_OF_MONTH))) // the day-of-week is valid
        {
            CAT.debug("yes");
            return true;
        }
        while (true)
        {
            // we switch to the next day
            CAT.debug("day index= "+event._currDayIdx+", day= "+event._days[event._currDayIdx]+", incrementing now");
            event._currDayIdx++;
            if (event._currDayIdx>=event._days.length||event._days[event._currDayIdx]>cal.getActualMaximum(Calendar.DAY_OF_MONTH))
            {
                // handling overflows
                CAT.debug("day index= "+event._currDayIdx+", day= "+event._days[event._currDayIdx]);
                event._currDayIdx=0;
                event._currMonthIdx++;
                CAT.debug("month index= "+event._currMonthIdx+", month= "+event._months[event._currMonthIdx]);
                if (event._currMonthIdx>=event._months.length)
                {
                    event._currMonthIdx=0;
                    event._currYearIdx++;
                    if (event._currYearIdx>=event._years.length)
                    {
                        CAT.debug("no - "+event); 
                        return false;
                    }
                    cal.set(Calendar.YEAR,event._years[event._currYearIdx]);
                }
                cal.set(Calendar.MONTH,event._months[event._currMonthIdx]);
            }
            cal.set(Calendar.DAY_OF_MONTH,event._days[event._currDayIdx]);
            dow=cal.get(Calendar.DAY_OF_WEEK);
            // check if this date has a valid day-of-week
            pos=Arrays.binarySearch(event._daysOfWeek,dow);
            if (0<=pos)
            {
                CAT.debug("yes - "+event); 
                return true;
            }
        }
    }
        
    protected long calculateTime(ActionEvent event)
    {
        Calendar cal=Calendar.getInstance();
        cal.setLenient(true);
        cal.set(event._years[event._currYearIdx],
                event._months[event._currMonthIdx],
                event._days[event._currDayIdx],
                event._hours[event._currHourIdx],
                event._minutes[event._currMinIdx],
                event._seconds[event._currSecIdx]
                );
        CAT.debug("scheduled time: "+event+" @"+cal.getTime());        
        return cal.getTime().getTime();
    }

    public void shutdown() throws java.lang.Exception
    {
        _shutDown=true;
        synchronized (_waitObj)
        {
            _waitObj.notifyAll();
        }
    }

    public void deInitialize() throws java.lang.Exception
    {
        _queue.clear();
    }

    public void reconfig(Node configNode) throws java.lang.Exception
    {
        try
        {
            NodeList nl=configNode.getChildNodes();
            for (int i=0;i<nl.getLength();i++)
            {
                Node node=nl.item(i);
                if (node.getNodeType()==node.ELEMENT_NODE)
                {
                    if(node.getNodeName().equals("event"))
                    {
                        addEvent(node);
                    }
                }
            }
        }
        catch (Exception ex)
        {
            CAT.error("erro during reconfiguration",ex);
        }
    }

    // processes the configuration for a single event
    protected void addEvent(Node configNode)
    {
        try
        {
            CAT.debug("addEvent()");
            ActionEvent event=new ActionEvent();

            NamedNodeMap attrs=configNode.getAttributes();
            event._id=attrs.getNamedItem("id").getNodeValue();

            NodeList nl=configNode.getChildNodes();
            for (int i=0;i<nl.getLength();i++)
            {
                Node node=nl.item(i);
                if (node.getNodeType()==node.ELEMENT_NODE)
                {
                    if(node.getNodeName().equals("action"))
                    {
                        NamedNodeMap nnl=node.getAttributes();
                        Node attr=nnl.getNamedItem("type");
                        String type=attr.getNodeValue();

                        if (type.equalsIgnoreCase("KILLSERVER"))
                        {
                            event._type=ActionEvent.KILLSERVER;
                        }
                        else if (type.equalsIgnoreCase("SHUTDOWN"))
                        {
                            event._type=ActionEvent.SHUTDOWN;
                        }
                        else if (type.equalsIgnoreCase("STARTUP"))
                        {
                            event._type=ActionEvent.STARTUP;
                        }
                        else if (type.equalsIgnoreCase("COMMAND"))
                        {
                            event._type=ActionEvent.COMMAND;
                        }

                        attr=nnl.getNamedItem("parameter");
                        if (null!=attr)
                            event._param=attr.getNodeValue();
                    }
                    else if(node.getNodeName().equals("time"))
                    {
                        NamedNodeMap nnl=node.getAttributes();
                        Node attr=nnl.getNamedItem("interval");
                        if (null!=attr)
                        {
                            String interval=attr.getNodeValue();
                            event._interval=Long.parseLong(interval);
                        }

                        attr=nnl.getNamedItem("year");
                        if (null!=attr)
                            event._year=attr.getNodeValue();

                        attr=nnl.getNamedItem("month");
                        if (null!=attr)
                            event._month=attr.getNodeValue();

                        attr=nnl.getNamedItem("day");
                        if (null!=attr)
                            event._day=attr.getNodeValue();

                        attr=nnl.getNamedItem("dayofweek");
                        if (null!=attr)
                            event._dayOfWeek=attr.getNodeValue();

                        attr=nnl.getNamedItem("hour");
                        if (null!=attr)
                            event._hour=attr.getNodeValue();

                        attr=nnl.getNamedItem("minute");
                        if (null!=attr)
                            event._minute=attr.getNodeValue();

                        attr=nnl.getNamedItem("second");
                        if (null!=attr)
                            event._second=attr.getNodeValue();

                        if (!event._day.equals("*")&&!event._dayOfWeek.equals("*"))
                        {
                            throw new InitializeException("day and day-of-week cannot be specified both, one must be '*'!");
                        }

                        if (event._second.equals("*"))
                        {
                            throw new InitializeException("'*' for second is not supported!");
                        }
                        
                        event._years=getValidEntries("year",event._year,2000,2010); // we assume that this software doesn't live _this_ long :)
                        event._months=getValidEntries("month",event._month,0,11);
                        event._days=getValidEntries("day",event._day,1,31);
                        event._daysOfWeek=getValidEntries("dow",event._dayOfWeek,1,7);
                        event._hours=getValidEntries("hour",event._hour,0,23);
                        event._minutes=getValidEntries("minute",event._minute,0,59);
                        event._seconds=getValidEntries("second",event._second,0,59);
                    }
                }
            }
            _queue.add(event);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    protected int[] getValidEntries(String fName, String field, int min, int max)
    {
        CAT.debug(fName+"="+field);
        Vector v=new Vector();
        if (field.equals("*"))
        {
            for (int i=min;i<=max;i++)
            {
                v.addElement(new Integer(i));
            }
        }
        else
        {
            boolean fini=false;
            while (!fini)
            {
                int pos=field.indexOf(',');
                String s=null;
                if (-1==pos)
                {
                    s=field;
                    fini=true;
                }
                else
                {
                    s=field.substring(0,pos);
                    field=field.substring(pos+1);
                }
                CAT.debug("subfield="+s);

                pos=s.indexOf('-');
                if (-1==pos)
                {
                    int i=Integer.parseInt(s);
                    v.addElement(new Integer(i));
                }
                else
                {
                    int to=Integer.parseInt(s.substring(pos+1));
                    int from=Integer.parseInt(s.substring(0,pos));
                    if (to>max)
                        to=max;
                    if (from<min)
                        from=min;
                    for (int i=from;i<=to;i++)
                    {
                        v.addElement(new Integer(i));
                    }
                }
            }
        }
        int[] ia=new int[v.size()];
        for (int i=0;i<ia.length;i++)
        {
//            CAT.debug("entry: "+((Integer)v.elementAt(i)).intValue());
            int j=((Integer)v.elementAt(i)).intValue();
            ia[i]=j;
        }
        // make sure the array is sorted
        Arrays.sort(ia);
        return ia;
    }

    protected void executeEvent(ActionEvent event)
    {
        switch (event._type)
        {
            case ActionEvent.KILLSERVER :
                    CAT.debug("KILLSERVER");
                    ProcessControlEvent pce=new ProcessControlEvent();
                    pce._type=ProcessControlEvent.KILL_PROCESS;
                    _bus.sendResponseData(pce,ProcessControlEvent.NAME,this);
                break;
            case ActionEvent.SHUTDOWN :
                    CAT.debug("SHUTDOWN");
                    pce=new ProcessControlEvent();
                    pce._type=ProcessControlEvent.SHUTDOWN;
                    _bus.sendResponseData(pce,ProcessControlEvent.NAME,this);
                break;
            case ActionEvent.STARTUP :
                    CAT.debug("STARTUP");
                    pce=new ProcessControlEvent();
                    pce._type=ProcessControlEvent.STARTUP;
                    _bus.sendResponseData(pce,ProcessControlEvent.NAME,this);
                break;
            case ActionEvent.COMMAND :
                    CAT.debug("COMMAND");
                    GameControlEvent gce=new GameControlEvent();
                    gce._type=GameControlEvent.USER_DEFINED;
                    gce._param1=event._param;
                    _bus.sendResponseData(gce,GameControlEvent.NAME,this);
                break;
            default:
                CAT.warn("can't execute action - unknown type ("+event._type+")");
        }
    }

    private class SchedulerThread extends Thread
    {
        public void run()
        {
            while (!_shutDown)
            {
                long sleepTime=-1;
                while (true)
                {
                    // take the time for every single run, so we are sure to have the correct time
                    // this is important when having many events which may take a time to process...
                    long now=System.currentTimeMillis();
                    CAT.debug("now=         "+now+" / "+new Date(now));
                    // if queue is empty, close thread...
                    if (_queue.isEmpty())
                        return;
                    ActionEvent event=(ActionEvent)_queue.first();
                    long startTime=event._nextRun;
                    CAT.debug("next run for "+event._id+"="+startTime+" / "+new Date(startTime));
                    sleepTime=startTime-now;
                    CAT.debug("sleepTime="+sleepTime);
                    // execute all events until 500msecs in the future
                    // events are then rescheduled
                    if (sleepTime<500)
                    {
                        CAT.debug("executing action for "+event._id);
                        executeEvent(event);
                        _queue.remove(event);
                        calculateNextRun(event);
                        CAT.debug("next run for "+event._id+"="+startTime);
                        if (-1!=event._nextRun)
                        {
                            _queue.add(event);
                        }
                    }
                    else
                    {
                        break;
                    }
                }
                CAT.debug("going to sleep");
                synchronized (_waitObj)
                {
                    try
                    {
                        _waitObj.wait(sleepTime);
                    }
                    catch (InterruptedException ex)
                    {
                        CAT.debug("got interrupted while sleeping - shutdown imminent?");
                    }
                }
            }
        }
    }

    private class ActionEvent implements Comparable
    {
        public static final int UNKNOWN=-1;
        public static final int KILLSERVER=2;
        public static final int SHUTDOWN=4;
        public static final int STARTUP=6;
        public static final int COMMAND=8;

        public String _id="none";

        public long _nextRun=-1;

        public long _interval=-1;

        public String _year="*";
        public String _month="*";
        public String _day="*";
        public String _dayOfWeek="*";
        public String _hour="*";
        public String _minute="*";
        public String _second="0";

        public int[] _years=null;
        public int[] _months=null;
        public int[] _days=null;
        public int[] _daysOfWeek=null;
        public int[] _hours=null;
        public int[] _minutes=null;
        public int[] _seconds=null;

        public int _currYearIdx=0;
        public int _currMonthIdx=0;
        public int _currDayIdx=0;
        public int _currHourIdx=0;
        public int _currMinIdx=0;
        public int _currSecIdx=0;

        public int _type=UNKNOWN;
        public String _param=null;

        public int compareTo(Object obj)
        {
            ActionEvent event=(ActionEvent)obj;
            if (_nextRun < event._nextRun)
                return -1;
            if (_nextRun > event._nextRun)
                return 1;
            // we cannot return 0 if the objext are not the same
            // otherwise we would lose events if 2 events have the same time...
            if (obj!=this)
                return 1;
            return 0;
        }

        public boolean equals(Object o)
        {
            return (o==this);
        }

        public String toString()
        {
            if (-1!=_interval)
            {
                return "interval alarm ("+_interval+" seconds) at " + new Date(_nextRun);
            }
            StringBuffer sb = new StringBuffer("Alarm params");
            sb.append(" second="); sb.append(_second);
            sb.append(" minute="); sb.append(_minute);
            sb.append(" hour="); sb.append(_hour);
            sb.append(" dayOfMonth="); sb.append(_day);
            sb.append(" month="); sb.append(_month);
            sb.append(" dayOfWeek="); sb.append(_dayOfWeek);
            sb.append(" year="); sb.append(_year);
            sb.append(" "); sb.append(_years[_currYearIdx]);
            sb.append("/"); sb.append(_months[_currMonthIdx]);
            sb.append("/"); sb.append(_days[_currDayIdx]);
            sb.append("/"); sb.append(_hours[_currHourIdx]);
            sb.append("/"); sb.append(_minutes[_currMinIdx]);
            sb.append("/"); sb.append(_seconds[_currSecIdx]);
            sb.append(" (next alarm=" + new Date(_nextRun) + ")");
            return sb.toString();
        }
    }
}