/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.control;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;

import org.log4j.Category;

import de.hendriklipka.util.*;

import de.hendriklipka.databus.*;

import de.hendriklipka.gcp.event.*;
import de.hendriklipka.gcp.exception.*;
import de.hendriklipka.gcp.module.*;

/**
    A QRunnerInputFilter starts a program and connects himself to Input / Output.
    StdOut + StdErr are converted into InputEvents.
    These InputEvents are broadcasted to the InfoBus. If OutputEvents are received
    from the InfoBus, they are sended to the running program.
*/

public class ServerRunner implements DataBusComponent,DataBusConsumer,DataBusProducer, Runnable
{
    static Category CAT = Category.getInstance(ServerRunner.class.getName());

    protected DataBus _bus=null;

    protected Process RunProcess=null;
    protected InputStream StdOut=null,StdErr=null;
    protected PrintStream StdIn=null;

    protected StringPipeOutput StdIn_Input=null;
    protected StringPipeInput StdIn_Output=null;
    protected Thread _processThread=null;

    protected String _command=null;
    protected int _level=0;
    protected boolean _restart=false;
    protected long _restartTime=0;
    protected boolean _detectHang=false;
    protected long _detectHangTime=0;
    protected long _lastOutput=0;
    protected boolean _doYield=false;

    protected boolean _shutDown=false;

    public void init(DataBus bus, Node configNode) throws Exception
    {
        _bus=bus;
        _bus.addConsumer(this,ProcessControlEvent.NAME);
        readConfig(configNode);
    }

    public void start()
    {
        if (!_shutDown)
        {
            CAT.debug("starting runner thread");
            Thread t=new Thread(this);
            t.setPriority(_level);
            _processThread=t;
            t.start();
        }
        else
        {
            CAT.debug("runner thread not started!");
        }
        _shutDown=false;
    }

    public void shutdown()
    {
        if (null!=_processThread)
        {
            _shutDown=true;
            if (null!=RunProcess)
            {
                RunProcess.destroy();
                RunProcess=null;
            }
        }
        else
        {
            CAT.debug("no thread for shutdown!");
        }
    }

    public void deInitialize()
    {
        _shutDown=true;
        if (null!=RunProcess)
        {
            RunProcess.destroy();
            RunProcess=null;
        }
    }

    public void reconfig(Node configNode) throws Exception
    {
        String cmd=_command;
        readConfig(configNode);
        // if we have a new command, we kill the old process and hope it is restarted :)
        if (!cmd.equals(_command))
        {
            Process oldProc=RunProcess;

            CAT.debug("starting after reconfig: <"+_command+">");
            RunProcess=Runtime.getRuntime().exec(_command);
            StdOut=RunProcess.getInputStream();
            StdErr=RunProcess.getErrorStream();
            StdIn=new PrintStream(RunProcess.getOutputStream());

            if (null!=oldProc)
            {
                CAT.debug("killing old process");
                oldProc.destroy();
            }
        }
    }

    public void readConfig(Node configNode)
    {
        String rlevel="+0";
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if (node.getNodeType()==node.ELEMENT_NODE)
            {
                if(node.getNodeName().equals("command"))
                {
                    _command=XMLUtil.getTextData(node);
                    NamedNodeMap attrs=node.getAttributes();
                    Node attr=attrs.getNamedItem("dontstart");
                    if (null!=attr)
                    {
                        // don't start the process now, but only at request
                        _shutDown=true;
                    }
                }
                else if(node.getNodeName().equals("priority"))
                {
                    rlevel=XMLUtil.getTextData(node);
                    CAT.debug("priority="+rlevel);
                }
                else if(node.getNodeName().equals("restart"))
                {
                    NamedNodeMap nnl=node.getAttributes();
                    Node attr=nnl.getNamedItem("time");
                    int l=Integer.parseInt(attr.getNodeValue());
                    if (-1==l)
                        _restart=false;
                    else
                        _restart=true;
                    _restartTime=l*1000;
                }
                else if(node.getNodeName().equals("detect_hang"))
                {
                    NamedNodeMap nnl=node.getAttributes();
                    Node attr=nnl.getNamedItem("time");
                    int l=Integer.parseInt(attr.getNodeValue());
                    if (-1==l)
                        _detectHang=false;
                    else
                        _detectHang=true;
                    _detectHangTime=l*1000;
                }
            }
        }

        _level=Thread.currentThread().NORM_PRIORITY;
        Integer I=null;
        if ('+'==rlevel.charAt(0))
        {
            rlevel=rlevel.substring(1);
            I=new Integer(rlevel);
            _level+=I.intValue();
        }
        else
        if ('-'==rlevel.charAt(0))
        {
            rlevel=rlevel.substring(1);
            I=new Integer(rlevel);
            _level-=I.intValue();
        }
        else
        {
            I=new Integer(rlevel);
            _level=I.intValue();
        }
        if (_level<Thread.MIN_PRIORITY)
            _level=Thread.MIN_PRIORITY;
        if (_level>Thread.MAX_PRIORITY)
            _level=Thread.MAX_PRIORITY;
        CAT.debug("final priority="+_level);
    }

    public boolean isGenericProducer()
    {
        return true;
    }

    public void run()
    {
        try
        {
            // start process
            try
            {
                CAT.debug("running: <"+_command+">");
                RunProcess=Runtime.getRuntime().exec(_command);
                StdOut=RunProcess.getInputStream();
                StdErr=RunProcess.getErrorStream();
                StdIn=new PrintStream(RunProcess.getOutputStream());
            }
            catch (Throwable ex)
            {
                CAT.error(ex);
            }

            long number=0;
            if (null==RunProcess)
            {
                CAT.error("no process");
                return;
            }
            boolean isRunning=true;
            while (isRunning&&!_shutDown)
            {
                // test, if the process is still running
                try
                {
                    // has the process finished?
                    // then we get an exception
                    int i=RunProcess.exitValue();
                    // we reach this point only if the process has finished
                    // process all output
                    // if there is output, we process this and do another round and continue the handling
                    if (0==StdOut.available())
                    {
                        if (!_restart)
                        {
                            isRunning=false;
                            CAT.debug("program has finished...");
                            return;
                        }
                        else
                        {
                            synchronized(this)
                            {
                            try
                            {
                                wait(_restartTime);
                            }
                            catch (InterruptedException ex) {}
                        }
                        CAT.debug("restarting: <"+_command+">");
                        RunProcess=Runtime.getRuntime().exec(_command);
                        StdOut=RunProcess.getInputStream();
                        StdErr=RunProcess.getErrorStream();
                        StdIn=new PrintStream(RunProcess.getOutputStream());
                        }
                    }
                }
                catch (IllegalThreadStateException ex) {}
                catch (IOException ex)
                {
                    CAT.error(ex);
                }
                catch (Throwable ex)
                {
                    CAT.error(ex);
                }
                try
                {
                    StringBuffer sb=new StringBuffer();
                    // transport StdIn
                    if (null!=StdIn_Output)
                    {
                        String s=StdIn_Output.read();
                        if ((null!=s)&&(null!=StdIn))
                        {
                            try
                            {
                                StdIn.println(s);
                                StdIn.flush();
                            }
                            catch (Exception ex)
                            {
                                CAT.error("StdInError: "+ex);
                            }
                        }
                    }
                    // send output of the process
                    int num=0;
                    if (0!=StdOut.available())
                    {
                        num++;
                        number++;
                        int b=' ';
                        while (b!='\n')
                        {
                            b=StdOut.read();
                            if ((b!='\n')&&(b!='\r'))
                                sb.append((char)b);
                        }
                        InputEvent ie=new InputEvent();
                        ie._eventDate=new Date();
                        ie._line=sb.toString();
                        _bus.sendData(ie,InputEvent.NAME,this);
                        Thread.currentThread().yield();
                    }
                    if (0!=StdErr.available())
                    {
                        num++;
                        number++;
                        int b=' ';
                        while (b!='\n')
                        {
                            b=StdErr.read();
                            if ((b!='\n')&&(b!='\r'))
                                sb.append((char)b);
                        }
                        InputEvent ie=new InputEvent();
                        ie._eventDate=new Date();
                        ie._line=sb.toString();
                        _bus.sendData(ie,InputEvent.NAME,this);
                    }
                    // if we have no output, we should have a short break, so we have not 100% CPU...
                    // if we have been told so, we also check if there was no input for s specified time
                    long time=System.currentTimeMillis();
                    if (0==num)
                    {
                        if (_detectHang)
                        {
                            if (0==_lastOutput)
                            {
                                _lastOutput=System.currentTimeMillis();
                            }
                            long diff=(time-_lastOutput);
                            if (diff>_detectHangTime)
                            {
                                // kill the process - restart will be handled at the top of the loop...
                                CAT.info("detected hanging process - killing <"+_command+">!");
                                RunProcess.destroy();
                                // set the last output time to current, to prevent infinite kill-restart orgies :)
                                _lastOutput=time;
                            }
                        }
                        synchronized (this)
                        {
                            wait(10);
                        }
                        continue;
                    }
                    else
                    {
                        _lastOutput=time;
                        if (_doYield)
                            Thread.currentThread().yield();
                    }
                    // ToDo - have a break every time??
                    /*
                    synchronized (this)
                    {
                    wait(10);
                    }
                    */
                }
                catch (Exception ex)
                {
                    CAT.error(ex);
                }
            } // while (isRunning)
            _shutDown=false;
            _processThread=null;
        }
        finally
        {
//            _bus.componentFinished(this);
        }
    }

    public synchronized void dataAvailable(de.hendriklipka.databus.DataBusEvent o)
    {
        try
        {
            ProcessControlEvent pce=(ProcessControlEvent)(o._data);
            if (pce._type==ProcessControlEvent.KILL_PROCESS)
            {
                CAT.info("got request - killing <"+_command+">!");
                RunProcess.destroy();
            }
            else if (pce._type==ProcessControlEvent.SHUTDOWN)
            {
                CAT.info("got request for shutdown");
                shutdown();
            }
            else if (pce._type==ProcessControlEvent.STARTUP)
            {
                CAT.info("got request for startup");
                if (null==_processThread)
                {
                    _shutDown=false;
                    start();
                }
                else
                {
                    CAT.debug("process already running");
                }
            }
        }
        catch (Throwable ex)
        {
            CAT.error(ex);
        }
    }

    public void preStartup() throws Exception
    {
    }

    public void postStartup() throws Exception
    {
    }
}
