/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.input;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;
import org.log4j.Category;

import de.hendriklipka.util.*;

import de.hendriklipka.databus.*;

import de.hendriklipka.gcp.event.*;
import de.hendriklipka.gcp.exception.*;
import de.hendriklipka.gcp.module.*;
import de.hendriklipka.gcp.util.*;

/**
    reads the input from a logfile, which is concurrently written by a program
    thus module is for getting the output from a Q3A server under windows, because this
    server does not print to StdOut
    the server should be set to write a log file and to flush after each line
    the file is periodically checked for net content
*/

public class LogFileInput implements DataBusComponent, DataBusProducer, Runnable
{
    static Category CAT = Category.getInstance(LogFileInput.class.getName());

    protected DataBus _bus=null;

    protected String _fileName=null;
    protected FileReader _inputReader=null;
    protected boolean _deleteFile=false;
    protected boolean _steadyListen=false;
    protected boolean _detectHang=false;
    protected long _detectHangTime=0;
    protected long _lastOutput=0;

    protected MessageTransformer _transformer=null;

    public void init(DataBus bus, Node configNode) throws Exception
    {
        _bus=bus;

        reconfig(configNode);
    }

    public void start()
    {
        Thread t=new Thread(this);
        t.start();
    }

    public void preStartup() throws Exception
    {
        if (_deleteFile)
        {
            File f=new File(_fileName);
            f.delete();
        }
    }

    public void shutdown()
    {
        _inputReader=null;
    }

    public void deInitialize()
    {
        _inputReader=null;
    }

    public boolean isGenericProducer()
    {
        return true;
    }

    public void run()
    {
        try
        {
            StringBuffer sb=new StringBuffer();
            while(_steadyListen||((null!=_inputReader)&&_inputReader.ready()))
            {
                try
                {
                    if (null==_inputReader)
                    {
                        if (null!=_inputReader)
                            _inputReader=new FileReader(_fileName);
                    }
                }
                catch (Exception ex)
                {
                }
                if ((null!=_inputReader)&&_inputReader.ready())
                {
                    int i=_inputReader.read();
                    if (-1!=i)
                    {
                        sb.append((char)i);
                        if ('\n'==i)
                        {
                            InputEvent ie=new InputEvent();
                            ie._originalLine=sb.toString();
                            if (null!=_transformer)
                            {
                                ie=_transformer.transform(ie);
                            }
                            else
                            {
                                ie._line=ie._originalLine;
                                ie._coloredLine=ie._line;
                            }
                            if (null==ie._eventDate)
                            {
                                ie._eventDate=new Date();
                            }
                            _bus.sendData(ie,InputEvent.NAME,this);
                            CAT.info(sb.toString());
                            sb=new StringBuffer();
                        }
                        // if there was new content, don't sleep and check for a new character
                        // but first we remember the current time...
                        _lastOutput=System.currentTimeMillis();
                    }
                }
                else
                {
                    // If there was no new output, check if this has happened for a longer period...
                    if (_detectHang)
                    {
                        if (0==_lastOutput)
                        {
                            _lastOutput=System.currentTimeMillis();
                        }
                        long time=System.currentTimeMillis();
                        long diff=time-_lastOutput;
                        if (diff>_detectHangTime)
                        {
                            // kill the process - restart will be handled at the top of the loop...
                            CAT.info("detected hanging process - sending KILL event!");
                            ProcessControlEvent pce=new ProcessControlEvent();
                            pce._type=pce.KILL_PROCESS;
                            _bus.sendData(pce,ProcessControlEvent.NAME,this);
                            _lastOutput=System.currentTimeMillis();
                        }
                    }
                    // if there was no new content, sleep a while
                    synchronized(this)
                    {
                        wait(100);
                    }
                }
            }
        }
        catch (Throwable ex)
        {
            CAT.error(ex);
        }
    }

    public void reconfig(Node configNode) throws Exception
    {
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if (node.getNodeType()==node.ELEMENT_NODE)
            {
                if(node.getNodeName().equals("inputfile"))
                {
                    _fileName=XMLUtil.getTextData(node);
                    NamedNodeMap attribs=node.getAttributes();
                    Node attr=attribs.getNamedItem("listen");
                    if (null!=attr)
                    {
                        String data=attr.getNodeValue().toLowerCase();
                        if (data.equals("1")||data.equals("yes")|data.equals("true"))
                        {
                            _steadyListen=true;
                        }
                    }
                    attr=attribs.getNamedItem("delete");
                    if (null!=attr)
                    {
                        String data=attr.getNodeValue().toLowerCase();
                        if (data.equals("1")||data.equals("yes")|data.equals("true"))
                        {
                            _deleteFile=true;
                        }
                    }
                }
                else if(node.getNodeName().equals("detect_hang"))
                {
                    NamedNodeMap nnl=node.getAttributes();
                    Node attr=nnl.getNamedItem("time");
                    int l=Integer.parseInt(attr.getNodeValue());
                    if (-1==l)
                        _detectHang=false;
                    else
                        _detectHang=true;
                    _detectHangTime=l*1000;
                }
                if(node.getNodeName().equals("transformer"))
                {
                    String cname=XMLUtil.getTextData(node);
                    if (null!=cname&&0!=cname.length())
                    {
                        try
                        {
                            Class c=Class.forName(cname);
                            Object o=c.newInstance();
                            _transformer=(MessageTransformer)o;
                        }
                        catch (Exception ex)
                        {
                            CAT.error(ex);
                        }
                    }
                }
            }
        }
    }

    public void postStartup() throws Exception
    {
    }

    public String[] produces()
    {
        String[] lt={InputEvent.NAME,ProcessControlEvent.NAME};
        return lt;
    }
}
