/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.kicker;

import org.w3c.dom.*;

import org.log4j.Category;

import de.hendriklipka.databus.*;

import de.hendriklipka.util.*;

import de.hendriklipka.gcp.event.*;

/**
    filters InputEvents for commands given with RCon commands with wrong password
    this is for implementing a command system for all users, not only admins, without touching the game
    such a command looks like:
    rcon $password $command $the_command $params
    the detection of the rcon command is done in the linefilter. this module generates
    specific messages for bad rcon commands (rcon commands with wrong passwords)
    - first, a message that there was a wrong rcon command
    - second, the wrong command itself
    
    $password and $command are configurable
    a new event with $the_command and $params is then created
    
    only tested with QW TF
    @see de.hendriklipka.gcp.module.processing.LineFilter
*/
public class RConUserCommandFilter implements DataBusComponent, DataBusConsumer, DataBusProducer
{
    static Category CAT = Category.getInstance(RConUserCommandFilter.class.getName());

    protected String _msgTypeBadRCon=null;
    protected String _msgTypeRConCommand=null;
    protected String _idBadRCon=null;
    protected String _idRConCommand=null;
    protected String _badRConIPProperty=null;
    protected String _rConCommandProperty=null;
    protected String _rConPasswordProperty=null;
    protected String _commandPrefix=null;
    
    protected String _userIP=null;
    
    protected int _state=NONE;
    
    protected DataBus _bus=null;
    
    protected static final int NONE=0;
    protected static final int RCON1=1;
    
    public void init(DataBus bus, Node configNode) throws Exception
    {
        _bus=bus;
        _bus.addConsumer(this,GameEvent.NAME);
        reconfig(configNode);        
    }

    public void dataAvailable(DataBusEvent event)
    {
        try
        {
            GameEvent ge=(GameEvent)(event._data);
            RConUserCommandEvent ruce=handleEvent(ge);
            if (null!=ruce)
            {
                _bus.sendData(ruce,RConUserCommandEvent.NAME,this);
            }
        }
        catch (Exception ex)
        {
            CAT.error(ex);
        }
    }

    protected RConUserCommandEvent handleEvent(GameEvent ge)
    {
        switch (_state)
        {
            case NONE:
                lookForBadRCon(ge);
                break;
            case RCON1:
                return lookForRConCommand(ge);
        }
        return null;
    }
    
    protected void lookForBadRCon(GameEvent ge)
    {
        if (null==_msgTypeBadRCon)
        {
            return;
        }
        if (!_msgTypeBadRCon.equals(ge._type))
        {
            return;
        }
        if ((null!=_idBadRCon)&&(!_idBadRCon.equals(ge._id)))
        {
            return;
        }
        _userIP=ge._data.getProperty(_badRConIPProperty);
        _state=RCON1;
    }
    
    protected RConUserCommandEvent lookForRConCommand(GameEvent ge)
    {
        if (null==_msgTypeRConCommand)
        {
            _state=NONE;
            return null;
        }
        if (!_msgTypeRConCommand.equals(ge._type))
        {
            _state=NONE;
            return null;
        }
        if ((null!=_idRConCommand)&&(!_idRConCommand.equals(ge._id)))
        {
            _state=NONE;
            return null;
        }
        String passwd=ge._data.getProperty(_rConPasswordProperty);
        String command=ge._data.getProperty(_rConCommandProperty);
        RConUserCommandEvent ruce=new RConUserCommandEvent();
        ruce._ip=_userIP;
        ruce._commandLine=command;
        ruce._password=passwd;
        if (null!=_commandPrefix)
        {
            if (command.startsWith(_commandPrefix))
            {
                command=command.substring(_commandPrefix.length());
                command=command.trim();
                ruce._command=command;
            }
            if (passwd.equals(_commandPrefix))
            {
                ruce._command=command;
            }
        }
        _state=NONE;
        return ruce;
    }
    
    public void deInitialize()
    {
    }

    public void start() throws Exception
    {
    }

    public void shutdown()
    {
    }

    public void reconfig(Node configNode)
    {
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if (node.getNodeType()==node.ELEMENT_NODE)
            {
                if (node.getNodeName().equals("badrcon_msgtype"))
                {
                    _msgTypeBadRCon=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("badrcon_msgid"))
                {
                    _idBadRCon=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("badrcon_property_ip"))
                {
                    _badRConIPProperty=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("rconcommand_msgtype"))
                {
                    _msgTypeRConCommand=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("rconcommand_msgid"))
                {
                    _idRConCommand=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("rconcommand_property_password"))
                {
                    _rConPasswordProperty=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("rconcommand_property_command"))
                {
                    _rConCommandProperty=XMLUtil.getTextData(node);
                }
                else if (node.getNodeName().equals("command_prefix"))
                {
                    _commandPrefix=XMLUtil.getTextData(node);
                }
            }
        }
    }

    public void preStartup() throws Exception
    {
    }
    
    public void postStartup() throws Exception
    {
    }
}
