/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
gcp@hendriklipka.de
*/

package de.hendriklipka.gcp.module.processing;

import java.util.*;
import java.io.*;

import org.w3c.dom.*;

import org.log4j.Category;

import de.hendriklipka.util.*;

import de.hendriklipka.databus.*;
import de.hendriklipka.gcp.module.control.*;
import de.hendriklipka.gcp.exception.*;
import de.hendriklipka.gcp.event.*;

/**
    this component converts InputEvents into GameEvents. This conversion works
    with a ruleset dependend on the possible strings in the InputEvents. So it is 
    independed from the type of the game or server.
    
    The possible input string are search until a match is found. This is then the resulting 
    GameEvent. GameEvents are in groups, every group is one type of GameEvent.
    
    Every string contains an unlimited number of fixed and variable parts. The
    following variable parts are allowed:
      
    %p# : name of known player
    %m# : name of known map
    %u# : name of unknown player. this name is then added to the list of known players
    %n# : number (double)
    %f# : filename
    %i# : IP-Adress (1.2.3.4 ... 255.255.255.255)
    %s# : string
    %c# : single character (Letter or Digit) (NOT at start of message)

    special substitions:
    %%  : %
    %"  : "

    the # is a unique number, which marks the variable part. the variable parts
    are mapped into named properties. this properties object is part of the
    resulting GameEvent.
    a message can start with:
    - a player name
    - a number
    - an IP address
    - a fixed string
    after every variable part must be a fixed part (except at the end)
*/

public class LineFilter implements DataBusComponent,DataBusConsumer,DataBusProducer
{
    static Category CAT = Category.getInstance(LineFilter.class.getName());

    protected DataBus _bus=null;
        
    protected char _pathSep=java.io.File.separatorChar;
      
    protected Hashtable _removedPlayers=new Hashtable();
      
    protected Vector    _playerByChar[]=new Vector[256];
    protected int[]     _existPlayer=new int[256];
    protected boolean[] _existMessage=new boolean[256];
    protected Vector[]  _messageByChar=new Vector[256];
    protected Vector    _messageWithPlayer=new Vector(10,10);
    protected Vector    _messageWithNumber=new Vector(10,10);
    protected Vector    _unknownMessage=new Vector(10,10);
    protected Hashtable _mapList=new Hashtable();
    
    protected String _unnamed="unnamed";
      
    protected long _lineCount=0;
    protected long _msgCount=0;

    // Player-Match-Modes
    protected final static int DIRECT_MATCH=1;
    protected final static int REMOVED_MATCH=2;
    protected final static int FUZZY_MATCH=3;
    protected int _playerMatchMode=REMOVED_MATCH;
    
    public void init(DataBus bus, Node configNode)
    {
        _bus=bus;
        _bus.addConsumer(this,InputEvent.NAME);
        
        reconfig(configNode);
    }
      
    /**
        get the GameEvent for this InputEvent
        
        strategy:
        - look at the first character
        - look at all messages which starts with a fixed part with this character
        - number-> look at all messages with %n / %i at the start
        - character -> if there is a player with this name -> look at all messages with %p
                    -> look at all messages with %u
    */
    protected GameEvent getGameEvent(InputEvent ie)
    {
        _lineCount++; // do a little bit of statistic work :)
        
        // we should have data to look at
        if (null==ie)
        {
            return null;
        }
        if (0==ie._line.length())
        {
            return null;
        }
        
        char c=ie._line.charAt(0);
        // if the message starts with a digit, look at all messages with a digit
        if ((true==Character.isDigit(c))||('-'==c))
        {
            GameEvent ge=findMessage(_messageWithNumber,ie._line);
            if (null!=ge)
            {
                postProcessGameEvent(ge,ie);
                return ge;
            }
        }
        
        // if the firstr character matches a known player, look at all player messages
        if (0!=_existPlayer[c])
        {
            if (_playerMatchMode>=DIRECT_MATCH)
            {
                GameEvent ge=findMessage(_messageWithPlayer,ie._line);
                if (null!=ge)
                {
                    postProcessGameEvent(ge,ie);
                    return ge;
                }
            }
        }
        
        // if me have messages which starts with this fices character, look at them
        if (true==_existMessage[c])
        {
            GameEvent ge=findMessage(_messageByChar[c],ie._line);
            if (null!=ge)
            {
                postProcessGameEvent(ge,ie);
                return ge;
            }
        }
        
        // if anything else fails, look at the rest...
        {
            GameEvent ge=findMessage(_unknownMessage,ie._line);
            if (null!=ge)
            {
                postProcessGameEvent(ge,ie);
                return ge;
            }
        }
        
        // no match - return an unknown message
        GameEvent ge=new GameEvent();
        ge._type="unknown";
        postProcessGameEvent(ge,ie);
        return ge;
    }

    /**
        tries to find a matching message for an input event
        if one is found, the replacement strings are put into a hashtable
    */
    protected GameEvent findMessage(Vector messages, String s)
    {
        for (int i=0;i<messages.size();i++)
        {
            ProcessData pd=new ProcessData();
            MessageInfo mi=(MessageInfo)(messages.elementAt(i));
            pd._msg=mi;
            if (existMessage(s,pd,0))
            {
                GameEvent ge=new GameEvent();
                ge._type=mi._type;
                // copy the mapped data to the GameEvent
                Enumeration e=mi._rules.keys();
                while (e.hasMoreElements())
                {
                    // name of property rule
                    String name=(String)e.nextElement();
                    // replacement string
                    String r=mi._rules.getProperty(name);
                    String rv=doReplace(r,pd._replaceData);
                    ge._data.put(name,rv);
                }
                ge._type=mi._type;
                handleSpecialMsg(pd);
                return ge;
            }
        }
        return null;
    }

    /**
        handles special messages like ADDPLAYER, REMOIVEPLAYER, STARTMAP and so on
    */    
    protected void handleSpecialMsg(ProcessData pd)
    {
        MessageInfo mi=pd._msg;
        if (null!=mi._addPlayer)
        {
            String s=mi._addPlayer;
            s=(String)pd._replaceData.get(s);
            addPlayerToList(s);
        }
        if (null!=mi._removePlayer)
        {
            String s=mi._removePlayer;
            s=(String)pd._replaceData.get(s);
            removePlayerFromList(s);
        }
        // ToDo
        // handle messages which are only available in certain maps
        // handle messages for startmap / endmap ...
    }
    
    protected String doReplace(String s, Hashtable p)
    {
        String res="";
        int pos=-1;
        while (true)
        {
            pos=s.indexOf('%');
            if (-1==pos)
            {
                res+=s;
                break;
            }
            res+=s.substring(0,pos);
            s=s.substring(pos);
            pos=s.indexOf(' ');
            String pn=null;
            if (-1==pos)
            {
                pn=s;
                s="";
            }
            else
            {
                pn=s.substring(0,pos);
                s=s.substring(pos);
            }
            String rv="";
            if (null!=p)
                rv=(String)p.get(pn);
            res+=rv;
        }
        return res;
    }
    
    protected void postProcessGameEvent(GameEvent ge, InputEvent ie)
    {
        ge._line=ie._line;
        ge._eventDate=ie._eventDate;
    }
    
    /**
        check, if the message <B>s</B> matches the one in <B>pd</B>
        this method is recursive
        
        @param s message to look at
        @param pd data for the message to compare with
        @param spos position in pd, where s must start

        strategy:
        look if the next part in <B>pd._msg</B> matches with the start of s
    */
    protected boolean existMessage(String s, ProcessData pd, int spos)
    {
        _msgCount++;
        String[] sa=pd._msg._asArray;
        String comp=s;
        int aPos=spos;
        while (true)
        {
            if (sa.length<=aPos)
                return true;
            String el=sa[aPos];
            char c=el.charAt(0);
            if ('%'!=c)
            {  // fixed string -> direct compare
                if (comp.startsWith(el))
                {
                    comp=comp.substring(el.length());
                    aPos++;
                    continue;
                }
                return false;
            }
            else
            {
                // handle %% and %"
                if (el.length()==2)
                {
                    c=el.charAt(1);
                    if (('%'==c)||('"'==c))
                    {
                        if (c==comp.charAt(0))
                        {
                            comp=comp.substring(1);
                            aPos++;
                            continue;
                        }
                    }
                }
                // now it is a variable substition
                if (aPos==(sa.length-1))
                {
                    // if there are no more elements, the string must fit completely to the variable part
                    return isSubstitute(el,comp,pd);
                }
                else
                {
                    // if there are more parts ->
                    // next part is a fixed one
                    // so wee look for all possible positions of this part in the string
                    // everything in between should match to the current variable part
                    // and the rest of the string should also match...
                    // we do this until we have a match or nothing is left
                    String el2=sa[aPos+1];
                    int pos=comp.indexOf(el2);
                    while (true)
                    {
                        if (-1==pos)
                        {
                            // no match, so we have no match...
                            return false;
                        }
                        String sub=comp.substring(0,pos);
                        if (isSubstitute(el,sub,pd))
                        {
                            // replacement
                            if ((comp.length()==(pos+el2.length()))&&(aPos==(sa.length-2)))
                            {
                                // el2 was the last one
                                return true;
                            }
                            else
                            {
                                return existMessage(comp.substring(pos),pd,aPos+1);
                            }
                        }
                        else
                        {
                            pos=comp.indexOf(el2,pos+1);
                        }
                    }
                }
            }
        }
    }

    /**
        check, if a %-part is represented by a string
        if yes, <B>pd</B> is updated
    */
    protected boolean isSubstitute(String element, String msgPart, ProcessData pd)
    {
        element=element.substring(1);
        char c1=element.charAt(0);
        switch (c1)
        {
            case 'u':
                // unknown players match everytime
                break;
            case 'p':
                if (!isStringAPlayer(msgPart))
                    return false;
                break;
            case 'n':
                if (!isStringANumber(msgPart))
                    return false;
                break;
            case 'm':
                if (!isStringAMap(msgPart))
                    return false;
                break;
            case 'f':
                if (!isStringAFile(msgPart))
                    return false;
                break;
            case 'i':
                if (!isStringAnIP(msgPart))
                    return false;
                break;
            case 's': // nothing to do, everything is a string :)
                break;
            case 'c': // everything with length==1 is a character
                if (1!=msgPart.length())
                    return false;
                break;
            default:
                return false;
        }
        if (null==pd._replaceData)
            pd._replaceData=new Hashtable(5);
        pd._replaceData.put("%"+element,msgPart);
        return true;
    }
    
    /**
        check, if a string represents a known player
        the strategy depends on the match mode:
        DIRECT_MATCH: the name must be in the list of known players
        REMOVED_MATCH: DIRECT_MATCH + the player could be in the list of currently removed players
        FUZZY_MATCH: fuzzy compare to find the player (after renaming ...)
    */
    protected boolean isStringAPlayer(String s)
    {
        if (0==s.length())
            return false;
        if (_playerMatchMode>=DIRECT_MATCH)
        {
            char c=s.charAt(0);
            Vector v=_playerByChar[c];
            for (int i=0;i<v.size();i++)
            {
                if (s.equals((String)v.elementAt(i)))
                    return true;
            }
        }
        if (_playerMatchMode>=REMOVED_MATCH)
        {
            if (null!=_removedPlayers.get(s))
            {
               _removedPlayers.remove(s);
                addPlayerToList(s);
                return true;
            }
        }
        // ToDo - create a real heuristic
        // in the moment we look only for an attached "(0)" (like QW does)
        if (_playerMatchMode>=FUZZY_MATCH)
        {
            char c=s.charAt(0);
            Vector v=_playerByChar[c];
            for (int i=0;i<v.size();i++)
            {
                String sv=(String)v.elementAt(i)+" (0)";
                if (s.equals(sv))
                    return true;
            }
        }
        if (s.equals(_unnamed))
            return true;
        return false;
    }
      
    /**
        filename:
        [DriveLetterr][Path]Filename
        DriveLetter=[a-zA-Z]:
        Path=Name PathSeparator [Path] 
        Filename=Name[.Ext]
        Name,Ext=[a-zA-Z][a-zA-Z_0-9]*
    */
    protected boolean isStringAFile(String s)
    {
        if (0==s.length())
        {
            return false;
        }
        if (1==s.length())
        {
            return isValidFileName(s);
        }
        if (':'==s.charAt(1))
        {
            if (Character.isLetter(s.charAt(0)))
                s=s.substring(2);
            else
            {
                return false;
            }
        }
        int pos=s.indexOf(_pathSep);
        while (-1!=pos)
        {
            if (0==pos)
            {
                s=s.substring(1);
                pos=s.indexOf(_pathSep);
                continue;
            }
            String s1=s.substring(0,pos);
            // string must be a valid filename
            if (!isValidFileName(s1))
            {
                return false;
            }
            s=s.substring(pos+1);
            pos=s.indexOf(_pathSep);
        }
        int i=s.indexOf('.');
        if (-1==i)
        {
            return isValidFileName(s);
        }
        String s1=s.substring(0,i);
        s=s.substring(i);
        return (isValidFileName(s)&&isValidFileName(s1));
    }
      
    protected boolean isValidFileName(String s)
    {
        for (int i=0;i<s.length();i++)
            if ((!Character.isLetter(s.charAt(i)))&&(!Character.isDigit(s.charAt(i)))&&('_'!=s.charAt(i))&&('.'!=s.charAt(i))&&('/'!=s.charAt(i)))
                return false;
        return true;
    }
      
    protected boolean isStringAMap(String s)
    {
        if (0==s.trim().length())
            return false;
        Object o=_mapList.get(s);
        return (null!=o);
    }
      
    protected boolean isStringAnIP(String s)
    {
        String n="";
        s=s.trim();
        Integer i=new Integer(0);
        int ii;
        try
        {
            // 1. number
            ii=s.indexOf('.');
            if (-1==ii)
                return false;
            n=s.substring(0,s.indexOf('.'));
            i=new Integer(n);
            ii=i.intValue();
            if ((0>ii)||(255<ii))
            {
                return false;
            }
            s=s.substring(s.indexOf('.'));
            if ('.'!=s.charAt(0))
            {
                return false;
            }
              
            // 2. number
            s=s.substring(1);
            ii=s.indexOf('.');
            if (-1==ii)
                return false;
            n=s.substring(0,s.indexOf('.'));
            i=new Integer(n);
            ii=i.intValue();
            if ((0>ii)||(255<ii))
            {
                return false;
            }
            s=s.substring(s.indexOf('.'));
            if ('.'!=s.charAt(0))
            {
                return false;
            }

            // 3. number
            ii=s.indexOf('.');
            if (-1==ii)
                return false;
            s=s.substring(1);
            n=s.substring(0,s.indexOf('.'));
            i=new Integer(n);
            ii=i.intValue();
            if ((0>ii)||(255<ii))
            {
                return false;
            }
            s=s.substring(s.indexOf('.'));
            if ('.'!=s.charAt(0))
            {
                return false;
            }
            // 4. number
            s=s.substring(1);
            ii=s.indexOf(':');
            if (-1!=ii)
                n=s.substring(0,ii);
            else
                n=s;
            i=new Integer(n);
            ii=i.intValue();
            if ((0>ii)||(255<ii))
            {
                return false;
            }
            return true;
        }
        catch (NumberFormatException ex)
        {
            return false;
        }
    }
    
    protected boolean isStringANumber(String s)
    {
        try
        {
            double d=(Double.valueOf(s)).doubleValue();
            return true;
        }
        catch (Exception ex)
        {
            return false;
        }
    }
      
    /**
        creates array of message parts
    */
    protected String[] messageToArray(String m)
    {
        char[] rep={'p','u','i','n','s','c','m','\'','%','f'};
        if (-1==m.indexOf('%'))
        {
            String[] sa=new String[1];
            sa[0]=m;
            return sa;
        }
        boolean fini=false;
        Vector v=new Vector(3,3);
        while (!fini)
        {
            int pos=m.indexOf('%');
            if (-1==pos)
            {
                if (0!=m.length())
                    v.addElement(m); 
                fini=true;
                break;
            }
            String s1=m.substring(0,pos);
            if (0!=s1.length())
                v.addElement(s1);
            m=m.substring(pos);
            if ('%'!=m.charAt(0))
            {
                continue;
            }
            for (int i=0;i<rep.length;i++)
            {
                if(rep[i]==m.charAt(1))
                {
                    StringBuffer sb=new StringBuffer();
                    sb.append('%');
                    sb.append(rep[i]);
                    i=2;
                    while ((i<m.length())&&(Character.isDigit(m.charAt(i))))
                    {
                        sb.append(m.charAt(i));
                        i++;
                    }
                    v.addElement(sb.toString());
                    m=m.substring(i);
                    break;
                }
            }
        }
        String[] saa=new String[v.size()];
        for (int i=0;i<v.size();i++)
        {
            saa[i]=(String)v.elementAt(i);
        }
        return saa;
    }
      
    /**
        adds a player to the internal list
    */
    protected void addPlayerToList(String n)
    {
        for (int i=0;i<_playerByChar[n.charAt(0)].size();i++)
            if (n.equals((String)_playerByChar[n.charAt(0)].elementAt(i)))
                return;
        _existPlayer[n.charAt(0)]++;
        _playerByChar[n.charAt(0)].addElement(n);
    }
      
    /**
        removes a player from the internal list and adds him to the "removed"-list
    */
    protected void removePlayerFromList(String n)
    {
        char c=n.charAt(0);
        Vector v=_playerByChar[c];
        for (int i=0;i<v.size();i++)
        {
            if (((String)v.elementAt(i)).equals(n))
            {
                _removedPlayers.put(n,new Object());
                v.removeElementAt(i);
                _existPlayer[c]--;
            }
        }
    }
      
    protected void removeAllPlayers()
    {
        for (int i=0;i<_playerByChar.length;i++)
        {
            for (int j=0;j<_playerByChar[i].size();j++)
            {
                _removedPlayers.put((String)(_playerByChar[i].elementAt(j)),new Object());
            }
            _playerByChar[i].removeAllElements(); 
        }
        addPlayerToList("unnamed");
    }
    
    protected void addMessage(MessageInfo messageInfo)
    {
        String[] ma=messageToArray(messageInfo._message);
        String m=messageInfo._message;
        messageInfo._asArray=ma;
        if ('%'==m.charAt(0))
        {
            if (m.substring(1,2).equals("p"))
            {
                _messageWithPlayer.addElement(messageInfo);
            }
            else if (m.substring(1,2).equals("u"))
            {
                _unknownMessage.addElement(messageInfo);
            }
            else if (m.substring(1,2).equals("i"))
            {
                _messageWithNumber.addElement(messageInfo);
            }
            else if (m.substring(1,2).equals("n"))
            {
                _messageWithNumber.addElement(messageInfo);
            }
            else if (m.substring(1,2).equals("%"))
            {
                _messageByChar['%'].addElement(messageInfo);
                _existMessage['%']=true;
            }
            else if (m.substring(1,2).equals("\""))
            {
                _messageByChar['"'].addElement(messageInfo);
                _existMessage['"']=true;
            }
            else
            {
                _existMessage[m.charAt(1)]=true;
                _messageByChar[m.charAt(1)].addElement(messageInfo);
            }
        }
        else
        {
            _existMessage[m.charAt(0)]=true;
            _messageByChar[m.charAt(0)].addElement(messageInfo);
        }
    }
      
    protected void addMap(String n)
    {
        _mapList.put(n,n);
    }
      
    /*
        for debugging
    */
    protected void DumpPlayer(PrintStream os)
    {
        os.println("--");
        for (int i=0;i<256;i++)
        {
            for (int j=0;j<_playerByChar[i].size();j++)
            {
                String s=(String)_playerByChar[i].elementAt(j);
                os.println("<"+s+">");
                //        os.print("<");
                //        for (int k=0;k<s.length();k++)
                //        {
                //          os.print("["+Integer.toHexString(s.charAt(k))+"]");
                //        }
                //        os.println(">");
            }
        }
        os.println("###");
    }
      
    public synchronized void dataAvailable(de.hendriklipka.databus.DataBusEvent o)
    {
        try
        {
            InputEvent ie=(InputEvent)(o._data);
            _playerMatchMode=DIRECT_MATCH;
            GameEvent ge=getGameEvent(ie);
            if (null==ge)
                return;
            if (null==ge._type)
            {
                _playerMatchMode=REMOVED_MATCH;
                ge=getGameEvent(ie);
                //ToDo: do this also for fuzzy match 
            }

            _bus.sendData(ge,GameEvent.NAME,this);
        }
        catch (Throwable ex)
        {
            CAT.error(ex);
        }
    }
      
    protected String doEntry(String s)
    {
        s=s.trim();
        if (0==s.length())
        {
            return "";
        }
        if ('"'==s.charAt(0))
            s=s.substring(1);
        if ('"'==s.charAt(s.length()-1))
            s=s.substring(0,s.length()-1);
        return s;
    }

    protected void readBlock(Node node, Properties props,String defType)
    {
        Properties myProps=(Properties)props.clone();
        String type=defType;
        NamedNodeMap nnl=node.getAttributes();
        Node attr=nnl.getNamedItem("type");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            type=id;
        }
        NodeList nl=node.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node thisNode=nl.item(i);
            if (thisNode.getNodeType()==thisNode.ELEMENT_NODE)
            {
                String name=thisNode.getNodeName();
                if (name.equals("msg"))
                {
                    readMessage(thisNode,myProps,type);
                }
                else if (name.equals("block"))
                {
                    readBlock(thisNode,myProps,type);
                }
                else if (name.equals("prop"))
                {
                    readProperty(thisNode, myProps);
                }
            }
        }
    }
    
    protected void readMessage(Node node, Properties props, String defType)
    {
        MessageInfo mi=new MessageInfo();
        mi._type=defType;
        Properties myProps=(Properties)props.clone();
        NamedNodeMap nnl=node.getAttributes();
        Node attr=nnl.getNamedItem("id");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            mi._id=id;
        }
        attr=nnl.getNamedItem("type");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            mi._type=id;
        }
        attr=nnl.getNamedItem("addplayer");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            mi._addPlayer=id;
        }
        attr=nnl.getNamedItem("removeplayer");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            mi._removePlayer=id;
        }
        attr=nnl.getNamedItem("startmap");
        if (null!=attr)
        {
            String id=attr.getNodeValue();
            mi._startMap=id;
        }
        attr=nnl.getNamedItem("endmap");
        if (null!=attr)
        {
            mi._endMap=true;
        }
        NodeList subs=node.getChildNodes();
        
        String s="";
        for (int j=0;j<subs.getLength();j++)
        {
            attr=subs.item(j);
            if (attr.getNodeType()==attr.ELEMENT_NODE)
            {
                if (attr.getNodeName().equals("prop"))
                {
                    readProperty(attr, myProps);
                }
                continue;
            }
            else if ((attr.getNodeType()==attr.TEXT_NODE)||(attr.CDATA_SECTION_NODE==attr.getNodeType()))
            {
                Text val=(Text)attr;
                s+=val.getData().trim();
            }
        }
        
        String line=doEntry(s);
        mi._message=line;
        mi._rules=myProps;
        addMessage(mi);
    }
    
    protected void readProperty(Node node, Properties props)
    {
        NamedNodeMap nnl=node.getAttributes();
        Node attr=nnl.getNamedItem("name");
        String name=attr.getNodeValue();

        attr=nnl.getNamedItem("value");
        String value=attr.getNodeValue();
        props.put(name,value);
    }
    
    private static class MessageInfo
    {
        public String _id=null;
        public String _message=null;
        public String[] _asArray=null;
        public String _type="---";
        public int _msgType=0;
        public final static int CHAR_START=1;
        public final static int PLAYER_START=2;
        public final static int NUM_START=3;
        public final static int STRING_START=4;
        public Properties _rules=new Properties();
        
        public String _removePlayer=null;
        public String _addPlayer=null;
        public String _startMap=null;
        public boolean _endMap=false;
    }
    
    private class ProcessData
    {
        public MessageInfo _msg=null; // message which is currently compared
        public Hashtable _replaceData=null; // contains the replaced variable parts
    }
    
    private class MapInfo
    {
        public String _id=null;
    }
    
    private class ItemInfo
    {
        public String _id=null;
    }
    
    private class ScoreInfo
    {
        public String _id=null;
    }
            
//------------------------------------------------

    public void start()
    {
    }
            
    public void shutdown()
    {
        CAT.info("Lines: "+_lineCount);
        CAT.info("Messages: "+_msgCount);
    }
            
    public void deInitialize()
    {
        CAT.info("Lines: "+_lineCount);
        CAT.info("Messages: "+_msgCount);
    }
            
    public void reconfig(Node configNode)
    {
        _existMessage=new boolean[256];
        _messageByChar=new Vector[256];
        _messageWithPlayer.removeAllElements();
        _messageWithNumber.removeAllElements();
        _unknownMessage.removeAllElements();
        _mapList.clear();
      
        for (int i=0;i<256;i++)
        {                                            
            _messageByChar[i]=new Vector(3,3);
            _playerByChar[i]=new Vector(3,3);                  
        }

        _existPlayer[_unnamed.charAt(0)]--;
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if (node.getNodeType()==node.ELEMENT_NODE)
            {
                if (node.getNodeName().equals("block"))
                {
                    readBlock(node,new Properties(),null);
                }
                if (node.getNodeName().equals("unnamedplayer"))
                {
                    _unnamed=XMLUtil.getTextData(node);
                    _existPlayer[_unnamed.charAt(0)]--;
                }
                if (node.getNodeName().equals("maps"))
                {
                    NodeList nl2=node.getChildNodes();
                    for (int j=0;j<nl2.getLength();j++)
                    {
                        Node node2=nl2.item(j);
                        if (node2.getNodeType()==node2.ELEMENT_NODE)
                        {
                            if (node2.getNodeName().equals("map"))
                            {
                                String mapName=XMLUtil.getTextData(node2);
                                addMap(mapName);
                            }
                        }
                    }
                }
            }
        }
    }
    
    public boolean isGenericProducer()
    {
        return false;
    }
        
    public void preStartup() throws Exception
    {
    }
    
    public void postStartup() throws Exception
    {
    }
}
